#include <stdio.h>
#include <stdint.h>
#include <time.h>
#include <mpi.h>
#include <gsl/gsl_rng.h>

/* https://www.wolframalpha.com/input/?i=lcm(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20)
 */
#define DEFAULT_STEPS 232792560ULL

/**
 *
 */
gsl_rng *new_random_generator(uint64_t seed)
{
  gsl_rng *random_generator;
  const gsl_rng_type *T;

  gsl_rng_env_setup();
  T = gsl_rng_default;
  random_generator = gsl_rng_alloc(T);
  gsl_rng_set(random_generator, seed);
  return random_generator;
}

/**
 *
 */
void free_random_generator(gsl_rng *random_generator)
{
  gsl_rng_free(random_generator);
}

/**
 *
 */
uint64_t compute_pi_montecarlo_hits(uint64_t steps, gsl_rng *random_generator)
{
  uint64_t i;
  uint64_t n_hits = 0;

  for (i = 0; i < steps; ++i) {
    double x = gsl_rng_uniform(random_generator);
    double y = gsl_rng_uniform(random_generator);
    double z = (x * x) + (y * y);
    if (z < 1.0) {
      ++n_hits;
    }
  }

  return n_hits;
}

/**
 *
 */
int main(int argc, char **argv)
{
  int comm_rank;
  int comm_size;
  gsl_rng *random_generator;
  double time_start;
  double time_end;
  uint64_t steps_by_process;
  uint64_t hits_by_process;
  uint64_t total_hits;
  double pi_montecarlo;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &comm_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &comm_size);

  random_generator = new_random_generator(time(0) + 1234ULL * comm_rank);
  time_start = MPI_Wtime();

  if (0 == comm_rank) {
    steps_by_process = DEFAULT_STEPS / comm_size;
  }

  MPI_Bcast(&steps_by_process, 1, MPI_INT64_T, 0, MPI_COMM_WORLD);

  hits_by_process = compute_pi_montecarlo_hits(steps_by_process, random_generator);

  total_hits = 0;
  MPI_Reduce(&hits_by_process, &total_hits, 1, MPI_INT64_T, MPI_SUM, 0, MPI_COMM_WORLD);

  time_end = MPI_Wtime();

  if (0 == comm_rank) {
    pi_montecarlo = (4.0 * total_hits) / (comm_size * steps_by_process);
    printf("Pi: %.8f\n", pi_montecarlo);
    printf("Elapsed time is %f\n", time_end - time_start);
  }

  free_random_generator(random_generator);
  MPI_Finalize();
  return 0;
}
